@@ -16,7 +16,10 @@
 #include "../mptrack/TrackerSettings.h"
 #include "../mptrack/Moddoc.h"
 #include "../mptrack/Mptrack.h"
+#include "mpt/io_file/inputfile.hpp"
+#include "mpt/io_file_read/inputfile_filecursor.hpp"
 #include "../common/mptFileIO.h"
+#include "mpt/fs/fs.hpp"
 #endif // MODPLUG_TRACKER
 
 OPENMPT_NAMESPACE_BEGIN
@@ -267,9 +270,12 @@ const char *szMidiPercussionNames[61] =
 };
 
 
+static constexpr uint8 NUM_MIDI_CHANNELS = 32;
+
+
 ////////////////////////////////////////////////////////////////////////////////
 // Maps a midi instrument - returns the instrument number in the file
-uint32 CSoundFile::MapMidiInstrument(uint8 program, uint16 bank, uint8 midiChannel, uint8 note, bool isXG, std::bitset<16> drumChns)
+uint32 CSoundFile::MapMidiInstrument(uint8 program, uint16 bank, uint8 midiChannel, uint8 note, bool isXG, std::bitset<NUM_MIDI_CHANNELS> drumChns)
 {
 	ModInstrument *pIns;
 	program &= 0x7F;
@@ -292,8 +298,9 @@ uint32 CSoundFile::MapMidiInstrument(uint8 program, uint16 bank, uint8 midiChann
 			if (program + 1 == p->nMidiProgram && bank + 1 == p->wMidiBank && p->nMidiDrumKey == 0) return i;
 		}
 	}
-	if ((m_nInstruments + 1 >= MAX_INSTRUMENTS) || (m_nSamples + 1 >= MAX_SAMPLES)) return 0;
-	
+	if(!CanAddMoreInstruments() || !CanAddMoreSamples())
+		return 0;
+
 	pIns = AllocateInstrument(m_nInstruments + 1);
 	if(pIns == nullptr)
 	{
@@ -304,9 +311,9 @@ uint32 CSoundFile::MapMidiInstrument(uint8 program, uint16 bank, uint8 midiChann
 	pIns->wMidiBank = bank + 1;
 	pIns->nMidiProgram = program + 1;
 	pIns->nFadeOut = 1024;
-	pIns->nNNA = NNA_NOTEOFF;
-	pIns->nDCT = isDrum ? DCT_SAMPLE : DCT_NOTE;
-	pIns->nDNA = DNA_NOTEFADE;
+	pIns->nNNA = NewNoteAction::NoteOff;
+	pIns->nDCT = isDrum ? DuplicateCheckType::Sample : DuplicateCheckType::Note;
+	pIns->nDNA = DuplicateNoteAction::NoteFade;
 	if(isDrum)
 	{
 		pIns->nMidiChannel = MIDI_DRUMCHANNEL;
@@ -327,13 +334,13 @@ uint32 CSoundFile::MapMidiInstrument(uint8 program, uint16 bank, uint8 midiChann
 	// Set GM program / drum name
 	if (!isDrum)
 	{
-		strcpy(pIns->name, szMidiProgramNames[program]);
+		pIns->name = szMidiProgramNames[program];
 	} else
 	{
 		if (note >= 24 && note <= 84)
-			strcpy(pIns->name, szMidiPercussionNames[note - 24]);
+			pIns->name = szMidiPercussionNames[note - 24];
 		else
-			strcpy(pIns->name, "Percussions");
+			pIns->name = "Percussions";
 	}
 	return m_nInstruments;
 }
@@ -350,81 +357,51 @@ struct MThd
 MPT_BINARY_STRUCT(MThd, 10)
 
 
-typedef uint32 tick_t;
+using tick_t = uint32;
 
 struct TrackState
 {
 	FileReader track;
-	tick_t nextEvent;
-	uint8 command;
-	bool finished;
-
-	TrackState()
-		: nextEvent(0)
-		, command(0)
-		, finished(false)
-	{ }
+	tick_t nextEvent = 0;
+	uint8 command = 0;
+	uint8 midiBaseChannel = 0;
+	bool finished = false;
 };
 
 struct ModChannelState
 {
-	static const uint8 NOMIDI = 0xFF;	// No MIDI channel assigned.
+	static constexpr uint8 NOMIDI = 0xFF;  // No MIDI channel assigned.
 
-	tick_t age;				// At which MIDI tick the channel was triggered
-	int32 porta;			// Current portamento position in extra-fine slide units (1/64th of a semitone)
-	uint8 vol;				// MIDI note volume (0...127)
-	uint8 pan;				// MIDI channel panning (0...256)
-	uint8 midiCh;			// MIDI channel that was last played on this channel
-	ModCommand::NOTE note;	// MIDI note that was last played on this channel
-	bool sustained : 1;		// If true, the note was already released by a note-off event, but sustain pedal CC is still active
-
-	ModChannelState()
-		: age(0)
-		, porta(0)
-		, vol(100)
-		, pan(128)
-		, midiCh(NOMIDI)
-		, note(NOTE_NONE)
-		, sustained(false)
-	{ }
+	tick_t age = 0;                     // At which MIDI tick the channel was triggered
+	int32 porta = 0;                    // Current portamento position in extra-fine slide units (1/64th of a semitone)
+	uint8 vol = 100;                    // MIDI note volume (0...127)
+	uint8 pan = 128;                    // MIDI channel panning (0...256)
+	uint8 midiCh = NOMIDI;              // MIDI channel that was last played on this channel
+	ModCommand::NOTE note = NOTE_NONE;  // MIDI note that was last played on this channel
+	bool sustained = false;             // If true, the note was already released by a note-off event, but sustain pedal CC is still active
 };
 
 struct MidiChannelState
 {
-	int32  pitchbendMod;	// Pre-computed pitchbend in extra-fine slide units (1/64th of a semitone)
-	int16  pitchbend;		// 0...16383
-	uint16 bank;			// 0...16383
-	uint8  program;			// 0...127
-	// -- Controllers ------------- function ---------- CC# --- range  ---- init (midi) ---
-	uint8 pan;             // Channel Panning           10      [0-255]     128  (64)
-	uint8 expression;      // Channel Expression        11      0-128       128  (127)
-	uint8 volume;          // Channel Volume            7       0-128       80   (100)
-	uint16 rpn;            // Currently selected RPN    100/101  n/a
-	uint8 pitchBendRange;  // Pitch Bend Range                              2
-	int8  transpose;       // Channel transpose                             0
-	bool  monoMode : 1;    // Mono/Poly operation       126/127  n/a        Poly
-	bool  sustain : 1;     // Sustain pedal             64       on/off     off
+	int32  pitchbendMod = 0;  // Pre-computed pitchbend in extra-fine slide units (1/64th of a semitone)
+	int16  pitchbend = MIDIEvents::pitchBendCentre; // 0...16383
+	uint16 bank = 0;          // 0...16383
+	uint8  program = 0;       // 0...127
+	// -- Controllers ---------------- function ---------- CC# --- range  ---- init (midi) ---
+	uint8 pan = 128;          // Channel Panning           10      [0-255]     128  (64)
+	uint8 expression = 128;   // Channel Expression        11      0-128       128  (127)
+	uint8 volume = 80;        // Channel Volume            7       0-128       80   (100)
+	uint16 rpn = 0x3FFF;      // Currently selected RPN    100/101  n/a
+	uint8 pitchBendRange = 2; // Pitch Bend Range                              2
+	int8  transpose = 0;      // Channel transpose                             0
+	bool  monoMode = false;   // Mono/Poly operation       126/127  n/a        Poly
+	bool  sustain = false;    // Sustain pedal             64       on/off     off
 
-	CHANNELINDEX noteOn[128];	// Value != CHANNELINDEX_INVALID: Note is active and mapped to mod channel in value
+	std::array<CHANNELINDEX, 128> noteOn;  // Value != CHANNELINDEX_INVALID: Note is active and mapped to mod channel in value
 
 	MidiChannelState()
-		: pitchbendMod(0)
-		, pitchbend(MIDIEvents::pitchBendCentre)
-		, bank(0)
-		, program(0)
-		, pan(128)
-		, expression(128)
-		, volume(80)
-		, rpn(0x3FFF)
-		, pitchBendRange(2)
-		, transpose(0)
-		, monoMode(false)
-		, sustain(false)
 	{
-		for(auto &note : noteOn)
-		{
-			note = CHANNELINDEX_INVALID;
-		}
+		noteOn.fill(CHANNELINDEX_INVALID);
 	}
 
 	void SetPitchbend(uint16 value)
@@ -450,31 +427,32 @@ struct MidiChannelState
 	{
 		switch(rpn)
 		{
-		case 0:	// Pitch Bend Range
+		case 0: // Pitch Bend Range
 			pitchBendRange = std::max(value, uint8(1));
 			SetPitchbend(pitchbend);
 			break;
-		case 2:	// Coarse Tune
+		case 2: // Coarse Tune
 			transpose = static_cast<int8>(value) - 64;
 			break;
 		}
 	}
 
-	uint8 GetRPN() const
+	void SetRPNRelative(int8 value)
 	{
 		switch(rpn)
 		{
-		case 0:	// Pitch Bend Range
-			return pitchBendRange;
-		case 2:	// Coarse Tune
-			return transpose;
+		case 0: // Pitch Bend Range
+			pitchBendRange = static_cast<uint8>(std::clamp(pitchBendRange + value, 1, 0x7F));
+			break;
+		case 2: // Coarse Tune
+			transpose = mpt::saturate_cast<int8>(transpose + value);
+			break;
 		}
-		return 0;
 	}
 };
 
 
-static CHANNELINDEX FindUnusedChannel(uint8 midiCh, ModCommand::NOTE note, const std::vector<ModChannelState> &channels, bool monoMode, PatternRow patRow)
+static CHANNELINDEX FindUnusedChannel(uint8 midiCh, ModCommand::NOTE note, const std::vector<ModChannelState> &channels, bool monoMode, mpt::span<ModCommand> patRow)
 {
 	for(size_t i = 0; i < channels.size(); i++)
 	{
@@ -489,10 +467,10 @@ static CHANNELINDEX FindUnusedChannel(uint8 midiCh, ModCommand::NOTE note, const
 	CHANNELINDEX anyFreeChannel = CHANNELINDEX_INVALID;
 
 	CHANNELINDEX oldsetMidiCh = CHANNELINDEX_INVALID;
-	tick_t oldestMidiChAge = Util::MaxValueOfType(oldestMidiChAge);
+	tick_t oldestMidiChAge = std::numeric_limits<decltype(oldestMidiChAge)>::max();
 
 	CHANNELINDEX oldestAnyCh = 0;
-	tick_t oldestAnyChAge = Util::MaxValueOfType(oldestAnyChAge);
+	tick_t oldestAnyChAge = std::numeric_limits<decltype(oldestAnyChAge)>::max();
 
 	for(size_t i = 0; i < channels.size(); i++)
 	{
@@ -532,7 +510,7 @@ static CHANNELINDEX FindUnusedChannel(uint8 midiCh, ModCommand::NOTE note, const
 }
 
 
-static void MIDINoteOff(MidiChannelState &midiChn, std::vector<ModChannelState> &modChnStatus, uint8 note, uint8 delay, PatternRow patRow, std::bitset<16> drumChns)
+static void MIDINoteOff(MidiChannelState &midiChn, std::vector<ModChannelState> &modChnStatus, uint8 note, uint8 delay, mpt::span<ModCommand> patRow, std::bitset<NUM_MIDI_CHANNELS> drumChns)
 {
 	CHANNELINDEX chn = midiChn.noteOn[note];
 	if(chn == CHANNELINDEX_INVALID)
@@ -586,11 +564,27 @@ static void EnterMIDIVolume(ModCommand &m, ModChannelState &modChn, const MidiCh
 }
 
 
+CSoundFile::ProbeResult CSoundFile::ProbeFileHeaderMID(MemoryFileReader file, const uint64 *pfilesize)
+{
+	MPT_UNREFERENCED_PARAMETER(pfilesize);
+	char magic[4];
+	file.ReadArray(magic);
+	if(!memcmp(magic, "MThd", 4))
+		return ProbeSuccess;
+
+	if(!memcmp(magic, "RIFF", 4) && file.Skip(4) && file.ReadMagic("RMID"))
+		return ProbeSuccess;
+
+	return ProbeFailure;
+}
+
+
 bool CSoundFile::ReadMID(FileReader &file, ModLoadingFlags loadFlags)
 {
 	file.Rewind();
 
 	// Microsoft MIDI files
+	bool isRIFF = false;
 	if(file.ReadMagic("RIFF"))
 	{
 		file.Skip(4);
@@ -611,6 +605,7 @@ bool CSoundFile::ReadMID(FileReader &file, ModLoadingFlags loadFlags)
 				file.Skip(length);
 			} else
 			{
+				isRIFF = true;
 				break;
 			}
 		} while(file.CanRead(8));
@@ -648,10 +643,14 @@ bool CSoundFile::ReadMID(FileReader &file, ModLoadingFlags loadFlags)
 	const ORDERINDEX MPT_MIDI_IMPORT_MAX_ORDERS = MAX_ORDERS;
 #endif
 
-	m_songArtist = MPT_USTRING("MIDI Conversion");
-	m_madeWithTracker = MPT_USTRING("Standard MIDI File");
-	SetMixLevels(mixLevels1_17RC3);
-	m_nTempoMode = tempoModeModern;
+	m_songArtist = U_("MIDI Conversion");
+	m_modFormat.formatName = U_("Standard MIDI File");
+	m_modFormat.type = isRIFF ? UL_("rmi") : UL_("mid");
+	m_modFormat.madeWithTracker = U_("Standard MIDI File");
+	m_modFormat.charset = mpt::Charset::ISO8859_1;
+
+	SetMixLevels(MixLevels::v1_17RC3);
+	m_nTempoMode = TempoMode::Modern;
 	m_SongFlags = SONG_LINEARSLIDES;
 	m_nDefaultTempo.Set(120);
 	m_nDefaultSpeed = ticksPerRow;
@@ -671,38 +670,65 @@ bool CSoundFile::ReadMID(FileReader &file, ModLoadingFlags loadFlags)
 		ppqn = 96;
 	Order().clear();
 
-	MidiChannelState midiChnStatus[16];
+	std::array<MidiChannelState, NUM_MIDI_CHANNELS> midiChnStatus;
 	const CHANNELINDEX tempoChannel = m_nChannels - 2, globalVolChannel = m_nChannels - 1;
 	const uint16 numTracks = fileHeader.numTracks;
 	std::vector<TrackState> tracks(numTracks);
 	std::vector<ModChannelState> modChnStatus(m_nChannels);
-	std::bitset<16> drumChns;
+	std::bitset<NUM_MIDI_CHANNELS> drumChns;
 	drumChns.set(MIDI_DRUMCHANNEL - 1);
+	drumChns.set(MIDI_DRUMCHANNEL + 15);
 
-	for(uint16 t = 0; t < numTracks; t++)
+	tick_t timeShift = 0;
+	for(auto &track : tracks)
 	{
 		if(!file.ReadMagic("MTrk"))
 			return false;
-		tracks[t].track = file.ReadChunk(file.ReadUint32BE());
+		track.track = file.ReadChunk(file.ReadUint32BE());
 		tick_t delta = 0;
-		tracks[t].track.ReadVarInt(delta);
-		tracks[t].nextEvent = delta;
+		track.track.ReadVarInt(delta);
+		// Work-around for some MID files that assume that negative deltas exist (they don't according to the standard)
+		if(delta > int32_max)
+			timeShift = std::max(static_cast<tick_t>(~delta  + 1), timeShift);
+		track.nextEvent = delta;
+	}
+	if(timeShift != 0)
+	{
+		for(auto &track : tracks)
+		{
+			if(track.nextEvent > int32_max)
+				track.nextEvent = timeShift - static_cast<tick_t>(~track.nextEvent + 1);
+			else
+				track.nextEvent += timeShift;
+		}
 	}
 
 	uint16 finishedTracks = 0;
 	PATTERNINDEX emptyPattern = PATTERNINDEX_INVALID;
-	ORDERINDEX lastOrd = 0;
-	ROWINDEX lastRow = 0;
+	ORDERINDEX lastOrd = 0, loopEndOrd = ORDERINDEX_INVALID;
+	ROWINDEX lastRow = 0, loopEndRow = ROWINDEX_INVALID;
 	ROWINDEX restartRow = ROWINDEX_INVALID;
 	int8 masterTranspose = 0;
 	bool isXG = false;
 	bool isEMIDI = false;
-	bool isType2 = (fileHeader.format == 2);
+	bool isEMIDILoop = false;
+	const bool isType2 = (fileHeader.format == 2);
+
+	const auto ModPositionFromTick = [&](const tick_t tick, const tick_t offset = 0)
+	{
+		tick_t modTicks = Util::muldivr_unsigned(tick, quantize * ticksPerRow, ppqn * 4u) - offset;
+
+		ORDERINDEX ord = static_cast<ORDERINDEX>((modTicks / ticksPerRow) / patternLen);
+		ROWINDEX row = (modTicks / ticksPerRow) % patternLen;
+		uint8 delay = static_cast<uint8>(modTicks % ticksPerRow);
+
+		return std::make_tuple(ord, row, delay);
+	};
 
 	while(finishedTracks < numTracks)
 	{
 		uint16 t = 0;
-		tick_t tick = Util::MaxValueOfType(tick);
+		tick_t tick = std::numeric_limits<decltype(tick)>::max();
 		for(uint16 track = 0; track < numTracks; track++)
 		{
 			if(!tracks[track].finished && tracks[track].nextEvent < tick)
@@ -715,11 +741,7 @@ bool CSoundFile::ReadMID(FileReader &file, ModLoadingFlags loadFlags)
 		}
 		FileReader &track = tracks[t].track;
 
-		tick_t modTicks = Util::muldivr_unsigned(tick, quantize * ticksPerRow, ppqn * 4u);
-
-		ORDERINDEX ord = static_cast<ORDERINDEX>((modTicks / ticksPerRow) / patternLen);
-		ROWINDEX row = (modTicks / ticksPerRow) % patternLen;
-		uint8 delay = static_cast<uint8>(modTicks % ticksPerRow);
+		const auto [ord, row, delay] = ModPositionFromTick(tick);
 
 		if(ord >= Order().GetLength())
 		{
@@ -752,7 +774,7 @@ bool CSoundFile::ReadMID(FileReader &file, ModLoadingFlags loadFlags)
 		}
 
 		PATTERNINDEX pat = Order()[ord];
-		PatternRow patRow = Patterns[pat].GetRow(row);
+		auto patRow = Patterns[pat].GetRow(row);
 
 		uint8 data1 = track.ReadUint8();
 		if(data1 == 0xFF)
@@ -765,11 +787,11 @@ bool CSoundFile::ReadMID(FileReader &file, ModLoadingFlags loadFlags)
 
 			switch(data1)
 			{
-			case 1:	// Text
-			case 2:	// Copyright
+			case 1: // Text
+			case 2: // Copyright
 				m_songMessage.Read(chunk, len, SongMessage::leAutodetect);
 				break;
-			case 3:	// Track Name
+			case 3: // Track Name
 				if(len > 0)
 				{
 					std::string s;
@@ -781,11 +803,11 @@ bool CSoundFile::ReadMID(FileReader &file, ModLoadingFlags loadFlags)
 						m_songName = s;
 				}
 				break;
-			case 4:	// Instrument
-			case 5:	// Lyric
+			case 4: // Instrument
+			case 5: // Lyric
 				break;
-			case 6:	// Marker
-			case 7:	// Cue point
+			case 6: // Marker
+			case 7: // Cue point
 				{
 					std::string s;
 					chunk.ReadString<mpt::String::maybeNullTerminated>(s, len);
@@ -794,20 +816,24 @@ bool CSoundFile::ReadMID(FileReader &file, ModLoadingFlags loadFlags)
 					{
 						Order().SetRestartPos(ord);
 						restartRow = row;
+					} else if(!mpt::CompareNoCaseAscii(s, "loopEnd"))
+					{
+						std::tie(loopEndOrd, loopEndRow, std::ignore) = ModPositionFromTick(tick, 1);
 					}
 				}
 				break;
-			case 8:	// Patch name
-			case 9:	// Port name
+			case 8: // Patch name
+			case 9: // Port name
 				break;
-			case 0x2F:	// End Of Track
+			case 0x21: // MIDI port
+				tracks[t].midiBaseChannel = chunk.ReadUint8() * 16u;
+				break;
+			case 0x2F: // End Of Track
 				tracks[t].finished = true;
 				break;
 			case 0x51: // Tempo
 				{
-					uint8 tempoRaw[3];
-					chunk.ReadArray(tempoRaw);
-					uint32 tempoInt = (tempoRaw[0] << 16) | (tempoRaw[1] << 8) | tempoRaw[2];
+					uint32 tempoInt = chunk.ReadUint24BE();
 					if(tempoInt == 0)
 						break;
 					TEMPO newTempo(60000000.0 / tempoInt);
@@ -817,27 +843,39 @@ bool CSoundFile::ReadMID(FileReader &file, ModLoadingFlags loadFlags)
 					} else if(newTempo != tempo)
 					{
 						patRow[tempoChannel].command = CMD_TEMPO;
-						patRow[tempoChannel].param = mpt::saturate_cast<ModCommand::PARAM>(std::max(32.0, Util::Round(newTempo.ToDouble())));
+						patRow[tempoChannel].param = mpt::saturate_round<ModCommand::PARAM>(std::max(32.0, newTempo.ToDouble()));
 					}
 					tempo = newTempo;
 				}
 				break;
+			case 0x7F: // Sequencer specific
+				{
+					// Yamaha MIDI port selection
+					uint32 data = chunk.ReadUint32BE();
+					if(chunk.LengthIs(4) && (data & 0xFFFFFF00) == 0x43000100)
+						tracks[t].midiBaseChannel = static_cast<uint8>((data & 0xFF) * 16u);
+				}
+				break;
 
 			default:
 				break;
 			}
 		} else
 		{
+			uint8 command = tracks[t].command;
 			if(data1 & 0x80)
 			{
-				// Command byte (if not present, repeat previous command byte)
-				tracks[t].command = data1;
+				// Command byte (if not present, use running status for channel messages)
+				command = data1;
 				if(data1 < 0xF0)
+				{
+					tracks[t].command = data1;
 					data1 = track.ReadUint8();
+				}
 			}
-			uint8 midiCh = tracks[t].command & 0x0F;
+			const uint8 midiCh = ((command & 0x0F) + tracks[t].midiBaseChannel) % NUM_MIDI_CHANNELS;
 
-			switch(tracks[t].command & 0xF0)
+			switch(command & 0xF0)
 			{
 			case 0x80: // Note Off
 			case 0x90: // Note On
@@ -845,7 +883,7 @@ bool CSoundFile::ReadMID(FileReader &file, ModLoadingFlags loadFlags)
 					data1 &= 0x7F;
 					ModCommand::NOTE note = static_cast<ModCommand::NOTE>(Clamp(data1 + NOTE_MIN, NOTE_MIN, NOTE_MAX));
 					uint8 data2 = track.ReadUint8();
-					if(data2 > 0 && (tracks[t].command & 0xF0) == 0x90)
+					if(data2 > 0 && (command & 0xF0) == 0x90)
 					{
 						// Note On
 						CHANNELINDEX chn = FindUnusedChannel(midiCh, note, modChnStatus, midiChnStatus[midiCh].monoMode, patRow);
@@ -907,7 +945,7 @@ bool CSoundFile::ReadMID(FileReader &file, ModLoadingFlags loadFlags)
 						midiChnStatus[midiCh].pan = data2 * 2u;
 						for(auto chn : midiChnStatus[midiCh].noteOn)
 						{
-							if(chn != CHANNELINDEX_INVALID)
+							if(chn != CHANNELINDEX_INVALID && modChnStatus[chn].pan != midiChnStatus[midiCh].pan)
 							{
 								if(Patterns[pat].WriteEffect(EffectWriter(CMD_PANNING8, midiChnStatus[midiCh].pan).Channel(chn).Row(row)))
 								{
@@ -970,7 +1008,7 @@ bool CSoundFile::ReadMID(FileReader &file, ModLoadingFlags loadFlags)
 
 					case MIDIEvents::MIDICC_DataButtonincrement:
 					case MIDIEvents::MIDICC_DataButtondecrement:
-						midiChnStatus[midiCh].SetRPN(midiChnStatus[midiCh].GetRPN() + ((data1 == MIDIEvents::MIDICC_DataButtonincrement) ? 1 : -1));
+						midiChnStatus[midiCh].SetRPNRelative((data1 == MIDIEvents::MIDICC_DataButtonincrement) ? 1 : -1);
 						break;
 
 					case MIDIEvents::MIDICC_NonRegisteredParameter_Fine:
@@ -993,13 +1031,31 @@ bool CSoundFile::ReadMID(FileReader &file, ModLoadingFlags loadFlags)
 
 					case 111:
 						// Non-standard MIDI loop point. May conflict with Apogee EMIDI CCs (110/111), which is why we also check if CC 110 is ever used.
-						if(data2 == 0)
+						if(data2 == 0 && !isEMIDI)
 						{
 							Order().SetRestartPos(ord);
 							restartRow = row;
 						}
 						break;
 
+					case 118:
+						// EMIDI Global Loop Start
+						isEMIDI = true;
+						isEMIDILoop = false;
+						Order().SetRestartPos(ord);
+						restartRow = row;
+						break;
+
+					case 119:
+						// EMIDI Global Loop End
+						if(data2 == 0x7F)
+						{
+							isEMIDILoop = true;
+							isEMIDI = true;
+							std::tie(loopEndOrd, loopEndRow, std::ignore) = ModPositionFromTick(tick, 1);
+						}
+						break;
+
 					case MIDIEvents::MIDICC_AllControllersOff:
 						midiChnStatus[midiCh].ResetAllControllers();
 						break;
@@ -1039,7 +1095,7 @@ bool CSoundFile::ReadMID(FileReader &file, ModLoadingFlags loadFlags)
 				midiChnStatus[midiCh].SetPitchbend(data1 | (track.ReadUint8() << 7));
 				break;
 			case 0xF0: // General / Immediate
-				switch(midiCh)
+				switch(command & 0x0F)
 				{
 				case MIDIEvents::sysExStart: // SysEx
 				case MIDIEvents::sysExEnd: // SysEx (continued)
@@ -1156,8 +1212,8 @@ bool CSoundFile::ReadMID(FileReader &file, ModLoadingFlags loadFlags)
 				}
 			}
 
-			m.command = static_cast<ModCommand::COMMAND>(diff < 0 ? CMD_PORTAMENTODOWN : CMD_PORTAMENTOUP);
-			int32 absDiff = mpt::abs(diff);
+			m.command = (diff < 0) ? CMD_PORTAMENTODOWN : CMD_PORTAMENTOUP;
+			int32 absDiff = std::abs(diff);
 			int32 realDiff = 0;
 			if(absDiff < 16)
 			{
@@ -1178,7 +1234,7 @@ bool CSoundFile::ReadMID(FileReader &file, ModLoadingFlags loadFlags)
 				m.param = static_cast<uint8>(absDiff);
 				realDiff = absDiff * 4 * (ticksPerRow - 1);
 			}
-			chnState.porta += realDiff * sgn(diff);
+			chnState.porta += realDiff * mpt::signum(diff);
 		}
 
 		tick_t delta = 0;
@@ -1193,25 +1249,36 @@ bool CSoundFile::ReadMID(FileReader &file, ModLoadingFlags loadFlags)
 			// Add another sub-song for type-2 files
 			if(isType2 && finishedTracks < numTracks)
 			{
-				if(Order.AddSequence(false) == SEQUENCEINDEX_INVALID)
+				if(Order.AddSequence() == SEQUENCEINDEX_INVALID)
 					break;
 				Order().clear();
 			}
 		}
 	}
 
+	if(isEMIDILoop)
+		isEMIDI = false;
+
 	if(isEMIDI)
 	{
 		Order().SetRestartPos(0);
 	}
 
-	if(Order().IsValidPat(lastOrd))
+	if(loopEndOrd == ORDERINDEX_INVALID)
+		loopEndOrd = lastOrd;
+	if(loopEndRow == ROWINDEX_INVALID)
+		loopEndRow = lastRow;
+
+	if(Order().IsValidPat(loopEndOrd))
 	{
-		PATTERNINDEX lastPat = Order()[lastOrd];
-		Patterns[lastPat].Resize(lastRow + 1);
+		PATTERNINDEX lastPat = Order()[loopEndOrd];
+		if(loopEndOrd == lastOrd)
+			Patterns[lastPat].Resize(loopEndRow + 1);
 		if(restartRow != ROWINDEX_INVALID && !isEMIDI)
 		{
-			Patterns[lastPat].WriteEffect(EffectWriter(CMD_PATTERNBREAK, mpt::saturate_cast<ModCommand::PARAM>(restartRow)).Row(lastRow));
+			Patterns[lastPat].WriteEffect(EffectWriter(CMD_PATTERNBREAK, mpt::saturate_cast<ModCommand::PARAM>(restartRow)).Row(loopEndRow));
+			if(ORDERINDEX restartPos = Order().GetRestartPos(); loopEndOrd != lastOrd || restartPos <= std::numeric_limits<ModCommand::PARAM>::max())
+				Patterns[lastPat].WriteEffect(EffectWriter(CMD_POSITIONJUMP, mpt::saturate_cast<ModCommand::PARAM>(restartPos)).Row(loopEndRow));
 		}
 	}
 	Order.SetSequence(0);
@@ -1228,11 +1295,11 @@ bool CSoundFile::ReadMID(FileReader &file, ModLoadingFlags loadFlags)
 		{
 			channels.push_back(i);
 			if(modChnStatus[i].midiCh != ModChannelState::NOMIDI)
-				sprintf(ChnSettings[i].szName, "MIDI Ch %u", 1 + modChnStatus[i].midiCh);
+				ChnSettings[i].szName = MPT_AFORMAT("MIDI Ch {}")(1 + modChnStatus[i].midiCh);
 			else if(i == tempoChannel)
-				strcpy(ChnSettings[i].szName, "Tempo");
+				ChnSettings[i].szName = "Tempo";
 			else if(i == globalVolChannel)
-				strcpy(ChnSettings[i].szName, "Global Volume");
+				ChnSettings[i].szName = "Global Volume";
 		}
 	}
 	if(channels.empty())
@@ -1242,7 +1309,7 @@ bool CSoundFile::ReadMID(FileReader &file, ModLoadingFlags loadFlags)
 	if(GetpModDoc() != nullptr)
 	{
 		// Keep MIDI channels in patterns neatly grouped
-		std::sort(channels.begin(), channels.end(), [&modChnStatus] (CHANNELINDEX c1, CHANNELINDEX c2) -> bool
+		std::sort(channels.begin(), channels.end(), [&modChnStatus] (CHANNELINDEX c1, CHANNELINDEX c2)
 		{
 			if(modChnStatus[c1].midiCh == modChnStatus[c2].midiCh)
 				return c1 < c2;
@@ -1252,130 +1319,87 @@ bool CSoundFile::ReadMID(FileReader &file, ModLoadingFlags loadFlags)
 		GetpModDoc()->m_ShowSavedialog = true;
 	}
 
-	std::shared_ptr<CDLSBank> pCachedBank, pEmbeddedBank;
-	mpt::PathString cachedBankFile;
+	std::unique_ptr<CDLSBank> cachedBank, embeddedBank;
 
-	if(CDLSBank::IsDLSBank(file.GetFileName()))
+	if(CDLSBank::IsDLSBank(file.GetOptionalFileName().value_or(P_(""))))
 	{
 		// Soundfont embedded in MIDI file
-		pEmbeddedBank = std::make_shared<CDLSBank>();
-		pEmbeddedBank->Open(file.GetFileName());
+		embeddedBank = std::make_unique<CDLSBank>();
+		embeddedBank->Open(file.GetOptionalFileName().value_or(P_("")));
 	} else
 	{
 		// Soundfont with same name as MIDI file
-		for(const auto &ext : { MPT_PATHSTRING(".sf2"), MPT_PATHSTRING(".sbk"), MPT_PATHSTRING(".dls") })
+		for(const auto &ext : { P_(".sf2"), P_(".sf3"), P_(".sf4"), P_(".sbk"), P_(".dls") })
 		{
-			mpt::PathString filename = file.GetFileName().ReplaceExt(ext);
-			if(filename.IsFile())
+			mpt::PathString filename = file.GetOptionalFileName().value_or(P_("")).ReplaceExtension(ext);
+			if(mpt::native_fs{}.is_file(filename))
 			{
-				pEmbeddedBank = std::make_shared<CDLSBank>();
-				if(pEmbeddedBank->Open(filename))
+				embeddedBank = std::make_unique<CDLSBank>();
+				if(embeddedBank->Open(filename))
 					break;
 			}
 		}
 	}
 	ChangeModTypeTo(MOD_TYPE_MPT);
-	MIDILIBSTRUCT &midiLib = CTrackApp::GetMidiLibrary();
-	// Scan Instruments
-	for (INSTRUMENTINDEX nIns = 1; nIns <= m_nInstruments; nIns++) if (Instruments[nIns])
+	const MidiLibrary &midiLib = CTrackApp::GetMidiLibrary();
+	mpt::PathString cachedBankName;
+	// Load Instruments
+	for (INSTRUMENTINDEX ins = 1; ins <= m_nInstruments; ins++) if (Instruments[ins])
 	{
-		mpt::PathString pszMidiMapName;
-		ModInstrument *pIns = Instruments[nIns];
-		uint32 nMidiCode;
-		bool embedded = false;
+		ModInstrument *pIns = Instruments[ins];
+		uint32 midiCode = 0;
+		if(pIns->nMidiChannel == MIDI_DRUMCHANNEL)
+			midiCode = 0x80 | (pIns->nMidiDrumKey & 0x7F);
+		else if(pIns->nMidiProgram)
+			midiCode = (pIns->nMidiProgram - 1) & 0x7F;
 
-		if (pIns->nMidiChannel == MIDI_DRUMCHANNEL)
-			nMidiCode = 0x80 | (pIns->nMidiDrumKey & 0x7F);
-		else
-			nMidiCode = pIns->nMidiProgram & 0x7F;
-		pszMidiMapName = midiLib.MidiMap[nMidiCode];
-		if (pEmbeddedBank)
+		if(embeddedBank && embeddedBank->FindAndExtract(*this, ins, midiCode >= 0x80))
 		{
-			uint32 nDlsIns = 0, nDrumRgn = 0;
-			uint32 nProgram = (pIns->nMidiProgram != 0) ? pIns->nMidiProgram - 1 : 0;
-			uint32 dwKey = (nMidiCode < 128) ? 0xFF : (nMidiCode & 0x7F);
-			if ((pEmbeddedBank->FindInstrument(	(nMidiCode >= 128),
-				((pIns->wMidiBank - 1) & 0x3FFF),
-				nProgram, dwKey, &nDlsIns))
-				|| (pEmbeddedBank->FindInstrument(	(nMidiCode >= 128),	0xFFFF,
-				(nMidiCode >= 128) ? 0xFF : nProgram,
-				dwKey, &nDlsIns)))
-			{
-				if (dwKey < 0x80) nDrumRgn = pEmbeddedBank->GetRegionFromKey(nDlsIns, dwKey);
-				if (pEmbeddedBank->ExtractInstrument(*this, nIns, nDlsIns, nDrumRgn))
-				{
-					pIns = Instruments[nIns]; // Reset pIns because ExtractInstrument may delete the previous value.
-					if ((dwKey >= 24) && (dwKey < 24 + MPT_ARRAY_COUNT(szMidiPercussionNames)))
-					{
-						mpt::String::CopyN(pIns->name, szMidiPercussionNames[dwKey - 24]);
-					}
-					embedded = true;
-				}
-				else
-					pIns = Instruments[nIns]; // Reset pIns because ExtractInstrument may delete the previous value.
-			}
+			continue;
 		}
-		if((!pszMidiMapName.empty()) && (!embedded))
-		{
-			// Load From DLS Bank
-			if (CDLSBank::IsDLSBank(pszMidiMapName))
-			{
-				std::shared_ptr<CDLSBank> pDLSBank = nullptr;
 
-				if ((pCachedBank) && (!mpt::PathString::CompareNoCase(cachedBankFile, pszMidiMapName)))
+		const mpt::PathString &midiMapName = midiLib[midiCode];
+		if(!midiMapName.empty())
+		{
+			// Load from DLS/SF2 Bank
+			if(CDLSBank::IsDLSBank(midiMapName))
+			{
+				CDLSBank *dlsBank = nullptr;
+				if(cachedBank != nullptr && !mpt::PathCompareNoCase(cachedBankName, midiMapName))
 				{
-					pDLSBank = pCachedBank;
+					dlsBank = cachedBank.get();
 				} else
 				{
-					pCachedBank = std::make_shared<CDLSBank>();
-					cachedBankFile = pszMidiMapName;
-					if (pCachedBank->Open(pszMidiMapName)) pDLSBank = pCachedBank;
+					cachedBank = std::make_unique<CDLSBank>();
+					cachedBankName = midiMapName;
+					if(cachedBank->Open(midiMapName)) dlsBank = cachedBank.get();
 				}
-				if (pDLSBank)
+				if(dlsBank)
 				{
-					uint32 nDlsIns = 0, nDrumRgn = 0;
-					uint32 nProgram = (pIns->nMidiProgram != 0) ? pIns->nMidiProgram - 1 : 0;
-					uint32 dwKey = (nMidiCode < 128) ? 0xFF : (nMidiCode & 0x7F);
-					if ((pDLSBank->FindInstrument(	(nMidiCode >= 128),
-						((pIns->wMidiBank - 1) & 0x3FFF),
-						nProgram, dwKey, &nDlsIns))
-						|| (pDLSBank->FindInstrument(	(nMidiCode >= 128), 0xFFFF,
-						(nMidiCode >= 128) ? 0xFF : nProgram,
-						dwKey, &nDlsIns)))
-					{
-						if (dwKey < 0x80) nDrumRgn = pDLSBank->GetRegionFromKey(nDlsIns, dwKey);
-						pDLSBank->ExtractInstrument(*this, nIns, nDlsIns, nDrumRgn);
-						pIns = Instruments[nIns]; // Reset pIns because ExtractInstrument may delete the previous value.
-						if ((dwKey >= 24) && (dwKey < 24 + MPT_ARRAY_COUNT(szMidiPercussionNames)))
-						{
-							mpt::String::CopyN(pIns->name, szMidiPercussionNames[dwKey - 24]);
-						}
-					}
+					dlsBank->FindAndExtract(*this, ins, midiCode >= 0x80);
 				}
 			} else
 			{
 				// Load from Instrument or Sample file
-				InputFile f;
-
-				if(f.Open(pszMidiMapName))
+				mpt::IO::InputFile f(midiMapName, SettingCacheCompleteFileBeforeLoading());
+				if(f.IsValid())
 				{
 					FileReader insFile = GetFileReader(f);
-					if(insFile.IsValid())
+					if(ReadInstrumentFromFile(ins, insFile, false))
 					{
-						ReadInstrumentFromFile(nIns, insFile, false);
-						mpt::PathString szName = pszMidiMapName.GetFullFileName();
-						pIns = Instruments[nIns];
-						if (!pIns->filename[0]) mpt::String::Copy(pIns->filename, szName.ToLocale().c_str());
-						if (!pIns->name[0])
+						mpt::PathString filename = midiMapName.GetFilename();
+						pIns = Instruments[ins];
+						if(!pIns->filename[0]) pIns->filename = filename.ToLocale();
+						if(!pIns->name[0])
 						{
-							if (nMidiCode < 128)
+							if(midiCode < 0x80)
 							{
-								mpt::String::CopyN(pIns->name, szMidiProgramNames[nMidiCode]);
+								pIns->name = szMidiProgramNames[midiCode];
 							} else
 							{
-								uint32 nKey = nMidiCode & 0x7F;
-								if (nKey >= 24)
-									mpt::String::CopyN(pIns->name, szMidiPercussionNames[nKey - 24]);
+								uint32 key = midiCode & 0x7F;
+								if((key >= 24) && (key < 24 + std::size(szMidiPercussionNames)))
+									pIns->name = szMidiPercussionNames[key - 24];
 							}
 						}
 					}
@@ -1384,7 +1408,6 @@ bool CSoundFile::ReadMID(FileReader &file, ModLoadingFlags loadFlags)
 		}
 	}
 #endif // MODPLUG_TRACKER
-
 	return true;
 }
 
